import pytest
import enum


class _ParseException(Exception):
    pass


def test_manager_state_updates(fs):
    from woger import BaseWorkspace, BaseAction

    @enum.unique
    class Action(BaseAction):
        load_raw_xml = 'load_raw_xml'
        parse_xml = 'parse_xml'

    class Workspace(BaseWorkspace):
        pass

    ws = Workspace('ws')
    load_tracker = ws.track(Action.load_raw_xml)
    with load_tracker:
        assert load_tracker.started()
        assert not load_tracker.finished()
        assert not load_tracker.failed()

    assert load_tracker.finished()
    assert not load_tracker.started()
    assert not load_tracker.failed()

    parse_tracker = ws.track(Action.parse_xml)
    with pytest.raises(_ParseException):
        with parse_tracker:
            assert parse_tracker.started()
            assert not parse_tracker.failed()
            assert not parse_tracker.finished()
            assert load_tracker.finished()
            raise _ParseException()

    assert parse_tracker.failed()
    assert not parse_tracker.finished()
    assert not parse_tracker.started()
    assert load_tracker.finished()
    assert not load_tracker.started()
    assert not load_tracker.failed()




from functools import total_ordering
from pathlib import Path

import os
from typing import Union

from .action_tracker import ActionTracker
from .base_action import BaseAction
from .state import WorkspaceState
from .workspace_meta import WorkspaceMeta


class NotDefinedError(Exception):
    pass


class BaseWorkspace(metaclass=WorkspaceMeta):
    """Base class for Workspace subclasses

    Examples
    ========
    .. include:: docs/examples/basic_workspace.rst

    """
    state_path = '.state.json'

    def __init__(self, path: Union[bytes, str]):
        """Creates a workspace

        Parameters
        ----------
        path: path-like
            Workspace path
        """
        self.path = Path(path)
        os.makedirs(str(self.path), exist_ok=True)

        self.state = WorkspaceState(self.state_path)

    def __repr__(self):
        return '<{}({})>'.format(self.__class__.__name__, self.path)

    @property
    def id(self):
        """Workspace id attribute

        Id is used to sort workspaces from oldest to latest
        """
        dirname = self.path.parts[-1]
        try:
            return int(dirname)
        except ValueError:
            return str(dirname)

    def track(self, action: BaseAction) -> ActionTracker:
        """Tracks the action inside the workspace

        Creates an ActionTracker object with action and this workspace as constructor parameters

        Parameters
        ----------
        action: subclass of BaseAction
            Action to track
        """
        return ActionTracker(action, self)

    def __eq__(self, other):
        return self.path == other.path


import warnings
from operator import attrgetter, itemgetter
from typing import List, Optional, Iterable

from .base_workspace import BaseWorkspace


class WorkspaceStorage:
    """Stores workspaces"""
    def __init__(self, iterable: Optional[Iterable[BaseWorkspace]]=None, limit: Optional[int]=None):
        if iterable is None:
            iterable = []
        elif not isinstance(iterable, list):
            iterable = list(iterable)

        if limit is None:
            limit = float('+inf')

        self.workspaces = [
            (ws.id, ws)
            for ws in sorted(iterable, key=attrgetter('id'))
        ]

        self.limit = limit

        if len(self.workspaces) > self.limit:
            warnings.warn('Workspace count is greater than limit')

    def __iter__(self):
        yield from self.keys()

    def keys(self):
        yield from (key for key, _ in self.workspaces)

    def values(self) -> List[BaseWorkspace]:
        yield from (value for _, value in self.workspaces)

    def items(self):
        yield from self.workspaces

    def __getitem__(self, item) -> Optional[BaseWorkspace]:
        """Gets workspace from storage by id"""
        workspaces = dict(self.workspaces)
        return workspaces[item] if item in workspaces else None

    def at(self, index) -> Optional[BaseWorkspace]:
        """Gets workspace from storage by index

        To get the oldest workspace use index 0
        To get the latest workspace use index -1
        """
        if index >= len(self.workspaces):
            return None

        pair = self.workspaces[index]
        return pair[1]

    def add(self, workspace: BaseWorkspace):
        """Add a workspace

        Parameters
        ----------
        workspace: subclass of BaseWorkspace
            Workspace to be added
        """
        if not isinstance(workspace, BaseWorkspace):
            raise ValueError('Value must be of type {}'.format(BaseWorkspace.__name__))

        pair = (workspace.id, workspace)

        workspaces = list(self.workspaces)
        workspaces.append(pair)
        workspaces = sorted(workspaces, key=itemgetter(0))
        self.workspaces = workspaces

    def __repr__(self):
        return '<WorkspaceStorage {}>'.format(list(self.keys()))

    def __len__(self):
        return len(self.workspaces)




from pathlib import Path


def _get_path_property(hidden_name):
    def _joined_path(self):
        child_path = getattr(self, hidden_name)
        return self.path / child_path

    return property(_joined_path)


class WorkspaceMeta(type):
    def __new__(mcs, cls_name, bases, attrib):
        mapping = {}

        for name, value in attrib.items():
            if name.startswith('_') or name in ['path', 'id', 'track', 'state']:
                mapping[name] = value
                continue

            hidden_name = '_' + name
            mapping[hidden_name] = Path(value)
            mapping[name] = _get_path_property(hidden_name)

        return super().__new__(mcs, cls_name, bases, mapping)



from setuptools import find_packages, setup


NAME = 'woger'
DESCRIPTION = 'Workspace manager library'
URL = 'https://gitlab.com/grhiabor/woger'
EMAIL = 'grihabor@mail.ru'
AUTHOR = 'Borodin Gregory'

REQUIRED = [

]

setup(
    name=NAME,
    version='1.0.0',
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
)
